/************************* MPEG-2 NBC Audio Decoder **************************
 *                                                                           *
 "This software module was originally developed by 
 Fraunhofer Gesellschaft IIS / University of Erlangen (UER) 
 and edited by Sony Corporation in the course of 
 development of the MPEG-2 NBC/MPEG-4 Audio standard ISO/IEC 13818-7, 
 14496-1,2 and 3. This software module is an implementation of a part of one or more 
 MPEG-2 NBC/MPEG-4 Audio tools as specified by the MPEG-2 NBC/MPEG-4 
 Audio standard. ISO/IEC  gives users of the MPEG-2 NBC/MPEG-4 Audio 
 standards free license to this software module or modifications thereof for use in 
 hardware or software products claiming conformance to the MPEG-2 NBC/MPEG-4
 Audio  standards. Those intending to use this software module in hardware or 
 software products are advised that this use may infringe existing patents. 
 The original developer of this software module and his/her company, the subsequent 
 editors and their companies, and ISO/IEC have no liability for use of this software 
 module or modifications thereof in an implementation. Copyright is not released for 
 non MPEG-2 NBC/MPEG-4 Audio conforming products.The original developer
 retains full right to use the code for his/her  own purpose, assign or donate the 
 code to a third party and to inhibit third party from using the code for non 
 MPEG-2 NBC/MPEG-4 Audio conforming products. This copyright notice must
 be included in all copies or derivative works." 
 Copyright(c)1996.
 *                                                                           *
 ****************************************************************************/

/* CREATED BY :  Bernhard Grill -- June-96  */

/*******************************************************************************************
 *
 * Master module for T/F based codecs
 *
 * Authors:
 * BG    Bernhard Grill, Fraunhofer Gesellshaft / University of Erlangen
 * HP    Heiko Purnhagen, Uni Hannover <purnhage@tnt.uni-hannover,de>
 * OS    Otto Schnurr, Motorola <schnurr@email.mot.com>
 * NI    Naoki Iwakami, NTT <iwakami@splab.hil.ntt.jp>
 * CL    Chuck Lueck, TI <lueck@ti.com>
 * ADD   Alberto Duenas, NDS <alberto@ndsuk.com>
 * JB    Jeremy Bennett, NDS <jbennett@ndsuk.com>
 * RG    Ralf Geiger, FhG/IIS
 *
 *
 * Changes:
 * fixed bit stream handling: Heiko Purnhagen, Uni Hannover, 960705 
 * 29-jul-96   OS   Modified source for 80 column width display.
 * 01-aug-96   OS   Integrated HQ source code module.
 *                  Integrated original TF mapping source code module.
 *                  Re-activated command line strings for run-time selection
 *                  of these modules.
 * 19-aug-96   BG   merge + inclusion of window switching / first short block support 
 * 21-aug-96   HP   renamed to enc_tf.c, adapted to new enc.h
 * 26-aug-96   HP   updated to vm0_tf_02a
 * HP 960826   spectra_index2sf_index alloc bug fix (still repeated alloc !!!)
 * 28-aug-96   NI   Integrated NTT's VQ coder.
 *                  Provided automatic window switching.
 * 08-jun-97   CL   Modified for use in MPEG-2 AAC technical report encoder software.
 *                  Removed support for medium windows and NTT VQ.  Added the use of
 *                  AAC fill elements for padding the frame when needed.  Added call
 *                  to WriteSCE for writing AAC single_channel_element.  Added
 *                  use of AACQuantInfo structure to reduce the number of parameters
 *                  required for some function calls.
 * 14-sep-97   CL   Integrated psychoacoustic module and prediction.
 * 24-oct-97   CL   Modified call to prediction to enable reset for short windows.
 *                  Integrated TNS.  Moved call to psycho model prior to SSR pre-
 *                  processing.  Modified call to psycho model to use look ahead
 *                  buffer for window switching decision.  Added window switching
 *                  control to WINDOW_SWITCH_MODE of FFT_PE_WINDOW_SWITCHING.  Moved
 *                  energy computation for short and long filterbank data prior to
 *                  TNS.  Added control of max_sfb and scalefactor_grouping to
 *                  EncTfFrame. 
 * 03-dec-97   ADD & JB   Added the concept of prediction reset.
 * 17-dec-97   CL   Increased the dimension on spectral_line_vector and
 *                  spectral_line_vector_for_gc to 2*block_size_samples since the mdct
 *                  function was using them as such.
 * 22-Jan-98   CL   Added multichannel support.  Added initial MS and IS joint coding
 *                  modules.
 * 07-apr-98   RG   Added encoding of LFE channel: window shape is sine, block type is
 *                  ONLY_LONG_WINDOW, spectral coefficients with index higher 12 are 
 *                  zero, TNS is disabled, prediction is disabled, relevant ratios 
 *                  from psycho model are fixed to -24dB, LFE-bitrate is reduced 
 *                  compared to other channels.
 * 27-May-98   MCF  Added support for independent coupling channel.
 * 24-July-98  MCF  Added support for dynamic range control.   
 *
 ******************************************************************************************/


/* Oct-2003  modified by RSY,I2R for MPEG-4 audio scalable lossless (SLS) coding */

#include <memory.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include "tf_main.h"
#include "bitstream.h"
#include "interface.h" 
#include "common.h"	/* common module */
#include "drc.h"
#include "enc.h"
#include "block.h"
#include "psych.h"
#include "aac_back_pred.h"
#include "assert.h"
#include "mc_enc.h"
#include "ms.h"
#include "is.h"

/* son_AACpp */
#include "sony_local.h"

#ifdef MPEG4V1
/* LTP */
#include "nok_ltp_enc.h"
#endif 

/* other aac */
#include "aac_qc.h"
#include "all.h"
#include "aac_se_enc.h"

#ifdef I2R_LOSSLESS
#include "lit_ll_en.h"
#include "lit_ms.h"
#include "int_mdct.h"

#define ROUND(x) ((x>=0)?(int)(x+0.5):(int)(x-0.5))
#define SQRT2 1.41421356
double pcm_scaling[4] = {1.0, 1.0, 1.0/16, 1.0/256};
double osf_scaling[5] = {0.0, 1.0, 1.0/SQRT2, 0.0, 1.0/2.0};
#endif

static int     block_size_samples = 1024;  /* nr of samples per block in one! audio channel */
static int     med_win_in_long    = 2;     
static int     short_win_in_long  = 8;
static int     bw_lines_long, bw_lines_med, bw_lines_short;  /* no of lines of the spectrum which are processed */
static int     start_line = 0;
static int     max_ch;    /* no of of audio channels */
static int	   nchans;
static int     debugLevel;
static double *spectral_line_vector[MAX_TIME_CHANNELS];
static double *residual_spectrum[MAX_TIME_CHANNELS];
static double *reconstructed_spectrum[MAX_TIME_CHANNELS];
static double *overlap_buffer[MAX_TIME_CHANNELS];
static double *DTimeSigBuf[MAX_TIME_CHANNELS];
static double *DTimeSigLookAheadBuf[MAX_TIME_CHANNELS];
#ifdef MPEG4V1
static double *time_buffer[MAX_TIME_CHANNELS];
static double *ltp_overlap_buffer[MAX_TIME_CHANNELS];
#endif

#ifdef I2R_LOSSLESS


static int *rec_spectrum[MAX_TIME_CHANNELS];

static int int_DTimeSigBuf[MAX_TIME_CHANNELS][MAX_OSF*1024];
static int int_DTimeSigLookAheadBuf[MAX_TIME_CHANNELS][MAX_OSF*1024];
static int *mdctTimeBuf[MAX_TIME_CHANNELS];
static int *mdctFreqBuf[MAX_TIME_CHANNELS];

static double CoreScaling;

static int *lossless_time_buffer[MAX_OSF*MAX_TIME_CHANNELS];
static double *lossless_overlap_buffer[MAX_OSF*MAX_TIME_CHANNELS];
static int *lossless_time_error[MAX_OSF*MAX_TIME_CHANNELS];
static double *p_prev_time_signal[MAX_OSF*MAX_TIME_CHANNELS];
static double *p_prev2_time_signal[MAX_OSF*MAX_TIME_CHANNELS];

static LIT_LL_Info ll_info[MAX_TIME_CHANNELS];

static int *quant_aac[MAX_TIME_CHANNELS];
static int *int_spectral_line_vector[MAX_TIME_CHANNELS];

#endif


/* static variables for the pre processing module */
/* son_NBCpp */
static int     max_band[MAX_TIME_CHANNELS];
static double *spectral_line_vector_for_gc[MAX_TIME_CHANNELS];
static double *DTimeSigLookAheadBufForGC[MAX_TIME_CHANNELS];
static double *DTimeSigPQFDelayCompensationBuf[MAX_TIME_CHANNELS];
static double **DBandSigLookAheadBuf[MAX_TIME_CHANNELS];
static double **DBandSigBufForGCAnalysis[MAX_TIME_CHANNELS];
static double **DBandSigWithGCBuf[MAX_TIME_CHANNELS];
static GAINC  **gainInfo[MAX_TIME_CHANNELS];

/* LT prediction */
#ifdef MPEG4V1
static double *nok_tmp_DTimeSigBuf[MAX_TIME_CHANNELS]; /* temporary fix to the buffer size problem. */
#endif

/* static variables used by the T/F mapping */
/*static enum WIN_SWITCH_MODE win_switch_mode = STATIC_SHORT; */          /* Only short windows */
/*static enum WIN_SWITCH_MODE win_switch_mode = STATIC_LONG; */      /* Only long windows */
/*static enum WIN_SWITCH_MODE win_switch_mode = LS_STARTSTOP_SEQUENCE ;*/   /* Start-stop windows */
/*static enum WIN_SWITCH_MODE win_switch_mode = LONG_SHORT_SEQUENCE;*/     /* Cycle through windows */
static enum WIN_SWITCH_MODE win_switch_mode = FFT_PE_WINDOW_SWITCHING;  /* PE switching */
static enum WINDOW_TYPE last_block_type[MAX_TIME_CHANNELS];
static enum WINDOW_TYPE desired_block_type[MAX_TIME_CHANNELS];
static enum WINDOW_TYPE next_desired_block_type[MAX_TIME_CHANNELS];
static enum WINDOW_TYPE block_type_lossless_only;
static enum WINDOW_TYPE block_type_current;
static enum CH_CONFIG        channel_configuration = CHC_MONO;     /* later f(encPara) */
static enum PP_MOD_SELECT    pp_select = NONE;                     /* latef f(encPara) */
static GC_DATA_SWITCH   gc_switch = GC_PRESENT; 
/*static enum GC_DATA_SWITCH   gc_switch = GC_NON_PRESENT;*/ 
static enum QC_MOD_SELECT    qc_select = AAC_QC;                   /* later f(encPara) */
static enum TF_MOD_SELECT    tf_select = MDCT_AAC;                 /* later f(encPara) */
static enum TRANSPORT_STREAM transport_stream =  LENINFO_TSTREAM;  /* later f(encPara) */
static short cplenabl = 0;					/* Enables coupling from the command line */
static short isenable = 1;                                      /* Enable intensity stereo */
static short msenable = 1;                                      /* Enable ms stereo */
static short tnsenable = 1;                                     /* Enable tns */

#ifdef I2R_LOSSLESS
static short coreenable = 1;
#endif

static enum AAC_PROFILE      profile = LOW;

/* Additional variables for AAC */
static int aacAllowScalefacs=1;              /* Allow AAC scalefactors to be nonconstant */
static TNS_INFO tnsInfo[MAX_TIME_CHANNELS];
#ifdef MPEG4V1
static NOK_LT_PRED_STATUS nok_lt_status[MAX_TIME_CHANNELS];
#endif
#ifdef BSAC
extern int codec_type;
#endif


AACQuantInfo quantInfo[MAX_TIME_CHANNELS];               /* Info structure for AAC quantization and coding */

/* Channel information */
Ch_Info channelInfo[MAX_TIME_CHANNELS];

/* AAC shorter windows 960-480-120 */
static int useShortWindows=0;  /* don't use shorter windows */

static int cplidx;
static CPL_VARS cvars;
static DSPshort chincpl[MAX_TIME_CHANNELS] = {1, 1, 1, 1, 1, 1};
static DSPfract cplmix[MAX_TIME_CHANNELS];
static DSPshort **cplexps;

static int sampling_rate;
int max_pred_sfb;

/********************************************************/
/* functions EncTfInfo() and EncTfFree()   HP 21-aug-96 */
/********************************************************/


/* EncTfInfo() */
/* Get info about t/f-based encoder core. */
char *EncTfInfo (
  FILE *helpStream)		/* in: print encPara help text to helpStream */
				/*     if helpStream not NULL */
				/* returns: core version string */
{
  if (helpStream != NULL) {
    fprintf(helpStream,
	    PROGVER "\n"
	    "encoder parameter string format:\n"
	    "possible options:\n"
            "-main      Main profile. (default)\n"
            "-lc        Low-complexity profile.\n"
            "-ssr       SSR profile.\n"
/* Moved to encoder.c, CL */
/*
#ifdef MPEG4V1
	    "-pns <start_band_for_perceptual_noise_subst>\n"
#endif
*/
	    "\n");
  }
  return PROGVER;
}


/* EncTfFree() */
/* Free memory allocated by t/f-based encoder core. */

void EncTfFree ()
{
}



/*****************************************************************************************
 ***
 *** Function: EncTfInit
 ***
 *** Purpose:  Initialize the T/F-part and the macro blocks of the T/F part of the VM
 ***
 *** Description:
 *** 
 ***
 *** Parameters:
 ***
 ***
 *** Return Value:
 ***
 *** **** MPEG-4 VM ****
 ***
 ****************************************************************************************/
#ifdef BSAC // BSAC-MC
enum {
SPEAKER_UNDEFINED						= 0x0,
SPEAKER_FRONT_LEFT_RIGHT    = 0x1,
SPEAKER_FRONT_CENTER				= 0x2,
SPEAKER_LOW_FREQUENCY       = 0x4,
SPEAKER_BACK_LEFT_RIGHT     = 0x8,
SPEAKER_BACK_CENTER         = 0x10,
SPEAKER_SIDE_LEFT_RIGHT     = 0x20,
SPEAKER_RESERVED            = 0x40
};

typedef struct {
int numElm;
int channelMask;
} channelMap;


#define MAX_CHANNEL_ELEMENT 10
typedef enum {
	CENTER_FRONT,		/* 1 */
	LR_FRONT,				/* 2 */
	REAR_SUR,				/* 1 */
	LR_SUR,					/* 2 */
	LFE,						/* 1 */
	LR_OUTSIDE,			/* 2 */	
	RESERVED 	
}CH_CFG ;

typedef enum _TNS_COMPLEXITY {
  NO_TNS,
  TNS_MAIN,
  TNS_LC,
  TNS_SSR,
#ifdef VERSION2
  TNS_LD_512,
  TNS_LD_480,
#endif
  TNS_SCAL,
  TNS_BSAC
} TNS_COMPLEXITY;        /* enum type for TNS complexity parameters */


struct channelElementData {
  TNS_COMPLEXITY        tns_select;
  int     ms_select;
  int     pns_sfb_start;
  int     aacAllowScalefacs;
  enum DC_FLAG aacSimulcast;
  enum WIN_SWITCH_MODE  win_switch_mode;
  enum WINDOW_SHAPE_ADAPT windowShapeAdapt;


  int     max_sfb;
//  MSInfo  msInfo;
	
	int startChIdx;
	int endChIdx;
	int chNum;
	int extension;
	CH_CFG ch_type;

  /* variables used by the T/F mapping */
//  WINDOW_SHAPE windowShapePrev[MAX_TIME_CHANNELS];
//  WINDOW_SEQUENCE windowSequencePrev[MAX_TIME_CHANNELS];


//  HANDLE_NTT_WINSW win_sw_module;

  int windowSequenceCounter;
  int windowShapeSeqCounter;

/* 20060107 */
//  SBR_CHAN *sbrChan[2];
};
typedef struct channelElementData * HANDLE_CHANNEL_ELEMENT_DATA;

typedef struct {
	int ch_elm_total;
	struct channelElementData chElmData[MAX_CHANNEL_ELEMENT];
} BSAC_EX_INFO;


channelMap chMapTbl[] = 
{
	{0, SPEAKER_UNDEFINED},
	{1, SPEAKER_FRONT_CENTER}, 
	{1, SPEAKER_FRONT_LEFT_RIGHT}, 
	{2, SPEAKER_FRONT_LEFT_RIGHT|SPEAKER_FRONT_CENTER}, 
	{3, SPEAKER_FRONT_LEFT_RIGHT|SPEAKER_FRONT_CENTER|SPEAKER_BACK_CENTER},
	{3, SPEAKER_FRONT_LEFT_RIGHT|SPEAKER_FRONT_CENTER|SPEAKER_BACK_LEFT_RIGHT},
	{4, SPEAKER_FRONT_LEFT_RIGHT|SPEAKER_FRONT_CENTER|SPEAKER_LOW_FREQUENCY|SPEAKER_BACK_LEFT_RIGHT},
	{0, SPEAKER_UNDEFINED},
	{5, SPEAKER_FRONT_LEFT_RIGHT|SPEAKER_FRONT_CENTER|SPEAKER_LOW_FREQUENCY|SPEAKER_BACK_LEFT_RIGHT|SPEAKER_SIDE_LEFT_RIGHT}
};

static BSAC_EX_INFO data_info;
static BSAC_EX_INFO* bsac_ex_info = &data_info;
int bsac_lfePresent;
#endif // BSAC-MC


void EncTfInit (
  int         numChannel,	 /* in: num audio channels */
  float       sampling_rate_f,   /* in: sampling frequancy [Hz] */         
  float       bit_rate_f,        /* in: bit rate [bit/sec] */
  char        *encPara,		 /* in: encoder parameter string */
  int         quantDebugLevel,   /* in : quantDebugLevel */
  int         bitReservInit,     /* in : number of bits in reservoir */
  int         varBitRate,        /* in : equals 1 if variable bit rate, else 0 */
  int         *frameNumSample,	 /* out: num samples per frame */
  int         *delayNumSample,	 /* out: encoder delay (num samples) */
  BsBitBuffer *bitHeader,        /* out: header for bit stream */
  int		  icplenabl,			 /* couple enable flag */
  int         isdisable,         /* intensity stereo enable */
  int         msdisable,         /* ms stereo enable */
  int         tnsdisable,        /* tns enable */
  int         pnsSfbStart        /* Start band for PNS */
#ifdef I2R_LOSSLESS
  ,int		  type_PCM
  ,int		  coredisable
  ,int        osf
#endif
  )
{

  BsBitStream *header_stream;
  int chanNum;
  int bit_rate = (int)(bit_rate_f+.5);

  char *p_ctmp;
  long superframe_blocks     = 1;        /* no of blocks in a superframe; later f(encPara) */
  debugLevel=  quantDebugLevel;
  cplenabl = icplenabl;
  nchans = numChannel + cplenabl;

  sampling_rate = (int)(sampling_rate_f+.5);

#if 0
  if (numChannel != 1)
    CommonExit(1,"EncTfInit: audio data has more the one channel (%d)",
	       numChannel);
#else
  if (nchans > MAX_TIME_CHANNELS)
    CommonExit(1,"EncTfInit: audio data has more than %d channels (%d)",
	       MAX_TIME_CHANNELS,numChannel);
#endif

  header_stream = BsOpenBufferWrite(bitHeader);

  /* Set defaults */
  profile = LOW;
  qc_select = AAC_QC;           
  tf_select = MDCT_AAC;          
  pp_select = NONE;               /* disable pre-processing */
  transport_stream = AAC_RAWDATA_TSTREAM;

  /* Check for pre-processing command line switch */
  if( strstr( encPara, "-ssr" ) ) {
    profile = SSR;
    pp_select = AAC_PP;           /* enable pre-processing */
    qc_select = AAC_QC;           /* disable prediction */
    transport_stream = AAC_RAWDATA_TSTREAM;
  }

  /* Check for low-complexity command line switch */
  if( strstr( encPara, "-lc" ) ) {
    profile = LOW;
    qc_select = AAC_QC;          /* disable prediction */
  }

  /* Disable is, ms, tns if desired */
  isenable = !isdisable;
  msenable = !msdisable;
  tnsenable = !tnsdisable;

/* Moved to encoder.c, CL */
/*
#ifdef MPEG4V1
  if( (p_ctmp=strstr( encPara, "-pns" )) ) {
    if( sscanf( p_ctmp+5, "%d",&pns_sfb_start ) == 0 ) {
      CommonExit( 1, "Encode: parameter of -pns switch not found" );
    }
    qc_select = AAC_QC;
 }
#endif
*/
/* Replaced with this, CL */
#ifdef MPEG4V1
  pns_sfb_start = pnsSfbStart;
  if (pns_sfb_start<=PNS_MAX_START) {
    qc_select = AAC_QC;
  }
#endif

  /* set the return values */
  max_ch = numChannel;
  {
      *frameNumSample = block_size_samples;
      *delayNumSample = 2*block_size_samples;
  }

  /* some global initializations */
  for (chanNum=0;chanNum<MAX_TIME_CHANNELS;chanNum++) {
    DTimeSigBuf[chanNum]            = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double));
    DTimeSigLookAheadBuf[chanNum]   = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double) );
    spectral_line_vector[chanNum]   = (double*)calloc(MAX_OSF*2*block_size_samples,sizeof(double) );
#ifdef MPEG4V1
    time_buffer[chanNum] = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double) );
    ltp_overlap_buffer[chanNum] = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double) );
    nok_lt_status[chanNum].delay =  (int*)calloc(MAX_SHORT_WINDOWS ,sizeof(int));
    nok_tmp_DTimeSigBuf[chanNum]  = (double*)calloc(MAX_OSF*2*block_size_samples, sizeof(double));
#endif
#ifdef I2R_LOSSLESS

    mdctTimeBuf[chanNum] = (int *)calloc(MAX_OSF*block_size_samples*2,sizeof(int));
	mdctFreqBuf[chanNum] = (int *)calloc(MAX_OSF*block_size_samples*2,sizeof(int));

	rec_spectrum[chanNum] = (int *)calloc(MAX_OSF*block_size_samples,sizeof(int));

    lossless_time_buffer[chanNum] = (int*)calloc(MAX_OSF*block_size_samples,sizeof(int) );
	lossless_time_error[chanNum] = (int*)calloc(MAX_OSF*block_size_samples,sizeof(int) );
    lossless_overlap_buffer[chanNum] = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double) );
	p_prev_time_signal[chanNum] = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double) );
	p_prev2_time_signal[chanNum] = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double) );
	int_spectral_line_vector[chanNum] = (int *)calloc(MAX_OSF*block_size_samples,sizeof(int) );
	quant_aac[chanNum] = (int *)calloc(MAX_OSF*block_size_samples,sizeof(int) );

	ll_info[chanNum].stream = (unsigned char *) calloc(MAX_OSF*1024*4,sizeof (unsigned char));
	ll_info[chanNum].type_PCM = type_PCM;
	
	coreenable = !coredisable;

#endif

    if (qc_select == AAC_PRED) {
      residual_spectrum[chanNum]    = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double) );
    }  

    reconstructed_spectrum[chanNum] = (double*)calloc(MAX_OSF*block_size_samples,sizeof(double) );
  }

  /* calculate max. audio bandwidth */
  {
    /* table: audio_bandwidth( bit_rate ) */
    static long bandwidth[8][2] = { {64000,20000},{56000,16000},{48000,14000},
                                    {40000,12000},{32000,9000},{24000,6000},
                                    {16000,3500},{-1,0} 
                                  };
    int i;

    i = 0;
    while( bandwidth[i][0] > bit_rate ) {
      if( bandwidth[i][0] < 0 ) {
        i--;
        break;
      }
      i++;
    }

    bw_lines_long  = bandwidth[i][1] * block_size_samples *2 / sampling_rate;
    bw_lines_med   = bandwidth[i][1] * block_size_samples *2 / sampling_rate / med_win_in_long;
    bw_lines_short = bandwidth[i][1] * block_size_samples *2 / sampling_rate / short_win_in_long;


    /* Use a minimal band-width for a bit rate <= 6 kbps. */
    start_line = 0;
    if ( bit_rate <= 6000 )
    {
      bw_lines_long = 140;
      start_line = 6;
    }

/* Init dynamic range parameters */

	initDynRng();
}
  

  /* initialize pre-processing */
  for (chanNum=0;chanNum<MAX_TIME_CHANNELS;chanNum++) {
    last_block_type[chanNum] = ONLY_LONG_WINDOW;
    desired_block_type[chanNum] = ONLY_LONG_WINDOW;
    block_type_lossless_only = ONLY_LONG_WINDOW;//ONLY_SHORT_WINDOW;//

    /* son_NBCpp */
    if (pp_select == AAC_PP)
      {
	int band;
	max_band[chanNum] = (NBANDS * (start_line + bw_lines_long - 1) /
			       block_size_samples); /* max_bands[] =0,1,2,3 */

	spectral_line_vector_for_gc[chanNum] = (double*)calloc(2*block_size_samples, sizeof(double));
	DTimeSigPQFDelayCompensationBuf[chanNum]=(double *)calloc(PQFDELAY, sizeof(double));
	DTimeSigLookAheadBufForGC[chanNum]=(double *)calloc(block_size_samples,sizeof(double));

	DBandSigLookAheadBuf[chanNum] = (double **)calloc(NBANDS, sizeof(double *));
	DBandSigBufForGCAnalysis[chanNum] = (double **)calloc(NBANDS, sizeof(double *));
	DBandSigWithGCBuf[chanNum] = (double **)calloc(NBANDS, sizeof(double *));
	gainInfo[chanNum] = (GAINC **)calloc(NBANDS, sizeof(GAINC *));
    
	for (band = 0; band < NBANDS; band++) {
	  DBandSigLookAheadBuf[chanNum][band]     = (double*)calloc(block_size_samples/NBANDS, sizeof(double));
	  DBandSigBufForGCAnalysis[chanNum][band] = (double*)calloc(block_size_samples/NBANDS*3, sizeof(double));
	  DBandSigWithGCBuf[chanNum][band]        = (double*)calloc(block_size_samples/NBANDS*2, sizeof(double));
	  gainInfo[chanNum][band]                 = (GAINC *)calloc(short_win_in_long,sizeof(GAINC));
	}
      }
  }

    cplexps = (DSPshort **)calloc(MAX_TIME_CHANNELS, sizeof(DSPshort *));
    for (chanNum=0;chanNum<MAX_TIME_CHANNELS;chanNum++)
      cplexps[chanNum] = (DSPshort *)calloc(NSFB_LONG,sizeof(DSPshort));
	cvars.cplexps = cplexps;
	cplidx = max_ch;

  /* initialize psychoacoustic module */
  EncTf_psycho_acoustic_init();

  /* initialize t/f mapping */
  for (chanNum=0;chanNum<MAX_TIME_CHANNELS;chanNum++) {
    overlap_buffer[chanNum] = (double*)calloc(sizeof(double), MAX_OSF*block_size_samples );
  }

  /* initialize spectrum processing */
  /* initialize quantization and coding */
  {
    /* init encode spectrum aac */
    tf_init_encode_spectrum_aac( quantDebugLevel);
  }

  /* Init TNS */
  
	  for (chanNum=0;chanNum<MAX_TIME_CHANNELS;chanNum++) {
		TnsInit(sampling_rate,profile,&tnsInfo[chanNum]);
		quantInfo[chanNum].tnsInfo = &tnsInfo[chanNum];         /* Set pointer to TNS data */
	}
  /* Get max_sfb_pred */
  max_pred_sfb = GetMaxPredSfb(sampling_rate);
  
  /* Set syntax debug level */
  InitAACSyntax(debugLevel);

#ifdef MPEG4V1
  /* Init LTP predictor */
  for (chanNum=0;chanNum<MAX_TIME_CHANNELS;chanNum++) {
    nok_init_lt_pred (&nok_lt_status[chanNum]);
    quantInfo[chanNum].ltpInfo = &nok_lt_status[chanNum];  /* Set pointer to LTP data */
  }
#endif

#ifdef BSAC  // BSAC-MC
{	
		int i;
		bsac_ex_info->ch_elm_total = chMapTbl[numChannel].numElm; 		

		for(i=0; i<bsac_ex_info->ch_elm_total; i++)
		{
			struct channelElementData *chData  = &bsac_ex_info->chElmData[i];			
//			chData->tns_select = data->tns_select;
//			chData->ms_select = data->ms_select;
//			chData->win_switch_mode = data->win_switch_mode;
//			chData->windowShapeAdapt = data->windowShapeAdapt;
//			chData->max_sfb = data->max_sfb;
			{
				int j,k;
//				chData->msInfo.ms_mask = data->msInfo.ms_mask;
				for (j=0; j<MAX_SHORT_WINDOWS; j++) {
					for (k=0; k<SFB_NUM_MAX; k++) {
//						chData->msInfo.ms_used[j][k] = 0;
					}
				}
			}
		}

		{
			int chMask = chMapTbl[numChannel].channelMask;
			int tempMask = 1;		
			int ext = 0;
			int start = 0;

			for(i=0; i<bsac_ex_info->ch_elm_total; )
			{
				switch(tempMask & chMask)
				{			
					case SPEAKER_FRONT_LEFT_RIGHT :
						bsac_ex_info->chElmData[i].chNum = 2;
						bsac_ex_info->chElmData[i].startChIdx = start;
						bsac_ex_info->chElmData[i].endChIdx = start+1;
						bsac_ex_info->chElmData[i].extension = ext;		
						bsac_ex_info->chElmData[i].ch_type = LR_FRONT;
						start += 2; i++; ext = 1;									
						break;
					case SPEAKER_FRONT_CENTER :
						bsac_ex_info->chElmData[i].chNum = 1;
						bsac_ex_info->chElmData[i].startChIdx = start;
						bsac_ex_info->chElmData[i].endChIdx = start;
						bsac_ex_info->chElmData[i].extension = ext;		
						bsac_ex_info->chElmData[i].ch_type = CENTER_FRONT;
						start += 1; i++; ext = 1;
						break;
					case SPEAKER_LOW_FREQUENCY :
						bsac_ex_info->chElmData[i].chNum = 1;
						bsac_ex_info->chElmData[i].startChIdx = start;
						bsac_ex_info->chElmData[i].endChIdx = start;
						bsac_ex_info->chElmData[i].extension = ext;			
						bsac_ex_info->chElmData[i].ch_type = CENTER_FRONT;// LFE should be considered again
						start += 1; i++; ext = 1;
						bsac_lfePresent = 1;
						break;
					case SPEAKER_BACK_LEFT_RIGHT :
						bsac_ex_info->chElmData[i].chNum = 2;
						bsac_ex_info->chElmData[i].startChIdx = start;
						bsac_ex_info->chElmData[i].endChIdx = start+1;
						bsac_ex_info->chElmData[i].extension = ext;		
						bsac_ex_info->chElmData[i].ch_type = LR_SUR;
						start += 2; i++; ext = 1;
						break;
					case SPEAKER_BACK_CENTER :		
						bsac_ex_info->chElmData[i].chNum = 1;
						bsac_ex_info->chElmData[i].startChIdx = start;
						bsac_ex_info->chElmData[i].endChIdx = start;
						bsac_ex_info->chElmData[i].extension = ext;		
						bsac_ex_info->chElmData[i].ch_type = REAR_SUR;
						start += 1; i++; ext = 1;
						break;
					case SPEAKER_SIDE_LEFT_RIGHT :
						bsac_ex_info->chElmData[i].chNum = 2;
						bsac_ex_info->chElmData[i].startChIdx = start;
						bsac_ex_info->chElmData[i].endChIdx = start+1;
						bsac_ex_info->chElmData[i].extension = ext;	
						bsac_ex_info->chElmData[i].ch_type = LR_OUTSIDE;
						start += 2; i++; ext = 1;
						break;
					default :
						break;
				}
				tempMask <<= 1;
			}
		}
	}
#endif // BSAC-MC
}


/*****************************************************************************************
 ***
 *** Function:    EncTfFrame
 ***
 *** Purpose:     processes a block of time signal input samples into a bitstream
 ***              based on T/F encoding 
 ***
 *** Description:
 *** 
 ***
 *** Parameters:
 ***
 ***
 *** Return Value:  returns the number of used bits
 ***
 *** **** MPEG-4 VM ****
 ***
 ****************************************************************************************/

void EncTfFrame (
		 float *p_time_signal[MAX_TIME_CHANNELS],   /* array of pointers to the modulo time 
													   signal input buffers for each channel */
		 BsBitBuffer  *fixed_bitstream,     /* bits before bitres size ptr */
#ifdef I2R_LOSSLESS
		BsBitBuffer *ll_bitstream,
#endif

		 int frameAvailNumBit,           	/* in: total num bits available for 
											   this frame (incl. bit reservoir) */
		 int frameNumBit,		            /* in: average num bits per frame */
                 int minBitsNoOverflow,     /* in: minimum number of bits needed to be used
                                               to avoid overflowing the bit reservoir */
		 int frameMaxNumBit,		        /* in: max num bits per frame */
		 int lfePresent,
		 double *icplmix,					/* independent coupling coefficients */
		 float *cplSampleBuf,
		 DRC_INTF *dvars_inpt
#ifdef I2R_LOSSLESS
		 ,long ll_bits_to_used
		 ,int osf
		 ,int msStereoIntMDCTMode
#endif
		 )
{
  int msStereoIntMDCT;
  BsBitStream *fixed_stream;
#ifdef I2R_LOSSLESS
  BsBitStream *ll_stream;
#endif
  BsBitStream *var_stream;
  BsBitStream *tmp_stream;
  BsBitBuffer *tmpBitBuf;
  BsBitStream *gc_stream[MAX_TIME_CHANNELS];  /* Bitstream pointer */
 
  int used_bits;
  int drcbits;
  int chanNum;
  int nfbwchans;

  /* Energy array (computed before prediction for long windows) */
  double energy[MAX_TIME_CHANNELS][MAX_SCFAC_BANDS];
#ifdef I2R_LOSSLESS
  double noise_energy[MAX_TIME_CHANNELS][MAX_SCFAC_BANDS];
#endif

  /* determine the function parameters used earlier:   HP 21-aug-96 */
  int          average_bits = frameNumBit;
  int          available_bitreservoir_bits = frameAvailNumBit-frameNumBit;

  /* actual amount of bits currently in the bit reservoir */
  /* it is the job of this module to determine 
     the no of bits to use in addition to average_block_bits
     max. available: average_block_bits + available_bitreservoir_bits */
  int max_bitreservoir_bits = frameMaxNumBit-frameNumBit;

  /* max. allowed amount of bits in the reservoir  (used to avoid padding bits) */
  long num_bits_available;

  /* the generated bitstreams are copied into the buffers designated by the */
  /* two pointers fixed_bitstream and bitres_bitstream */
  BsBitBuffer  *bitres_bitstream = NULL;            /* bits after bitres size ptr */

  double *p_ratio[MAX_TIME_CHANNELS], allowed_distortion[MAX_TIME_CHANNELS][MAX_SCFAC_BANDS];
  static double p_ratio_long[2][MAX_TIME_CHANNELS][MAX_SCFAC_BANDS];
  static double p_ratio_short[2][MAX_TIME_CHANNELS][MAX_SCFAC_BANDS];
  int    nr_of_sfb[MAX_TIME_CHANNELS], sfb_width_table[MAX_TIME_CHANNELS][MAX_SCFAC_BANDS];
  int sfb_offset_table[MAX_TIME_CHANNELS][MAX_SCFAC_BANDS+1];
  static enum WINDOW_TYPE block_type[MAX_TIME_CHANNELS];

  int   win, no_sub_win, sub_win_size, bw_lines;
  
  /* structures holding the output of the psychoacoustic model */
  CH_PSYCH_OUTPUT chpo_long[MAX_TIME_CHANNELS];
  CH_PSYCH_OUTPUT chpo_short[MAX_TIME_CHANNELS][MAX_SHORT_WINDOWS];
  static int psychNum=0;
  psychNum = !psychNum;      /* Toggle psychNum, psycho model using look-ahead buffer */

  if (fixed_bitstream == NULL)
    return;	/* quick'n'dirty fix for encoder startup    HP 21-aug-96 */


  fixed_stream = BsOpenBufferWrite(fixed_bitstream);
#ifdef I2R_LOSSLESS
    ll_stream = BsOpenBufferWrite(ll_bitstream);
#endif
  if (bitres_bitstream != NULL)
    var_stream = BsOpenBufferWrite(bitres_bitstream);
  else
    var_stream = NULL;
  tmpBitBuf = BsAllocBuffer(32768);
  tmp_stream = BsOpenBufferWrite(tmpBitBuf);

  /* son_NBCpp */
  for (chanNum=0;chanNum<MAX_TIME_CHANNELS;chanNum++) {
    quantInfo[chanNum].gcBitBuf = BsAllocBuffer(4096);   /* Allocate bit buffer in quantInfo */
    gc_stream[chanNum] = BsOpenBufferWrite(quantInfo[chanNum].gcBitBuf);  /* Use bit buffer in quantInfo */
  }

  { /* convert float input to double, which is the internal format */
    /* store input data in look ahead buffer which may be necessary for the window switching decision */
   int i;
   int chanNum;
   
   for (chanNum=0;chanNum<max_ch;chanNum++) {
     for( i=0; i<osf*block_size_samples; i++ ) {
#ifdef MPEG4V1
         /* temporary fix: a linear buffer for LTP containing the whole time frame */
         nok_tmp_DTimeSigBuf[chanNum][i] = DTimeSigBuf[chanNum][i];
         nok_tmp_DTimeSigBuf[chanNum][block_size_samples + i] = DTimeSigLookAheadBuf[chanNum][i];
#endif
       /* last frame input data are encoded now */
       DTimeSigBuf[chanNum][i]          = DTimeSigLookAheadBuf[chanNum][i];

       /* YT 970615 for son_gc */
       if (pp_select == AAC_PP) {
         DTimeSigLookAheadBuf[chanNum][i] =  
	   (i < PQFDELAY) ? DTimeSigPQFDelayCompensationBuf[chanNum][i] : (double)p_time_signal[chanNum][i-PQFDELAY] ;
         DTimeSigLookAheadBufForGC[chanNum][i] = 
	   (double)p_time_signal[chanNum][i];
       }

       else { 
         DTimeSigLookAheadBuf[chanNum][i] = (double)p_time_signal[chanNum][i];
       } /* end if(pp_select ...) */
     } /* end for(i ..) */
   } /* end for(chanNum ... ) */

   if (cplenabl)
   {
	  lfePresent ? (nfbwchans = max_ch - 1) : (nfbwchans = max_ch);
	  for (chanNum = 0; chanNum < nfbwchans; chanNum++)
	  {
	 	 cplmix[chanNum] = icplmix[chanNum];
	  }
	  couple_e(cplmix, nfbwchans, cplSampleBuf);				/* independent coupling */

	  for( i=0; i<block_size_samples; i++ ) 
	  {
        /* last frame input data are encoded now */
        DTimeSigBuf[max_ch][i]          = DTimeSigLookAheadBuf[max_ch][i];

        /* YT 970615 for son_gc */
        if (pp_select == AAC_PP) {
         DTimeSigLookAheadBuf[max_ch][i] =  
	   /*	    (i < PQFDELAY) ? DTimeSigPQFDelayCompensationBuf[max_ch][i] : (double)p_time_signal[max_ch][i-PQFDELAY] ; */
	    (i < PQFDELAY) ? DTimeSigPQFDelayCompensationBuf[max_ch][i] : (double)cplSampleBuf[i-PQFDELAY] ;
          DTimeSigLookAheadBufForGC[max_ch][i] = (double)cplSampleBuf[i];
        }

        else { 
          DTimeSigLookAheadBuf[max_ch][i] = (double)cplSampleBuf[i];
        } /* end if(pp_select ...) */
	 }
   }
	
  /* Keep track of number of bits used */
   used_bits = 0;

	/* YT 960615 for son_pp */
   if (pp_select == AAC_PP) {
     for (chanNum=0;chanNum<max_ch;chanNum++) {
       for (i = 0; i < PQFDELAY; i++) {
	 DTimeSigPQFDelayCompensationBuf[chanNum][i] = 
	   (double)p_time_signal[chanNum][block_size_samples-PQFDELAY+i];
       }
     }
     if (cplenabl) {
	 DTimeSigPQFDelayCompensationBuf[max_ch][i] = 
	   (double)cplSampleBuf[block_size_samples-PQFDELAY+i];
     }       
   } 
  }

   /***********************************************************************/
   /* Determine channel elements      */
   /***********************************************************************/
   DetermineChInfo(channelInfo, nchans, cplenabl, lfePresent);



  /******************************************************************************************************************************
  *
  * psychoacoustic
  *
  ******************************************************************************************************************************/
#ifdef I2R_LOSSLESS
   if (coreenable)
   {
#endif
  {
    int chanNum;
    for (chanNum=0;chanNum<nchans;chanNum++) {
      EncTf_psycho_acoustic( sampling_rate, chanNum, &DTimeSigLookAheadBuf[chanNum], &next_desired_block_type[chanNum],
			     (int)qc_select,block_size_samples ,&chpo_long[chanNum], &chpo_short[chanNum] );

      if (channelInfo[chanNum].lfe) {        /* dummy psychoacoustic for LFE */
        int sfb;
        for (sfb=0;sfb<4;sfb++) { 
          chpo_long[chanNum].p_ratio[sfb]=0.004; /* -24 dB */
        } 
        next_desired_block_type[chanNum] = ONLY_LONG_WINDOW;
      }
    }
  }

#ifdef I2R_LOSSLESS
   }
   else
	   for (chanNum=0;chanNum<nchans;chanNum++) {
   {
	   EncTf_settable(sampling_rate,&chpo_long[chanNum]);
	   EncTf_settable_short(sampling_rate,&chpo_short[chanNum]);        

   }
   }
#endif

  /******************************************************************************************************************************
  *
  * block_switch processing 
  *
  ******************************************************************************************************************************/

  {


    int chanNum;

#ifdef I2R_LOSSLESS

    win_switch_mode = STATIC_LONG;		// I2R todo: check for lossless decoding using window switching

#endif

    for (chanNum=0;chanNum<nchans;chanNum++) {

      if (channelInfo[chanNum].lfe) {
	if (debugLevel>2)
	  printf("WSM: ONLY_LONG_WINDOW\n");
	block_type[chanNum] = ONLY_LONG_WINDOW;   /* For LFE only long windows */
      } else { 
        switch( win_switch_mode ) {
        case STATIC_LONG:
          if (debugLevel>2)
	    printf("WSM: ONLY_LONG_WINDOW\n");
  	  block_type[chanNum] = ONLY_LONG_WINDOW;
 	  break;
        case STATIC_SHORT :
	  if (debugLevel>2)
	    printf("WSM:ONLY_SHORT_WINDOW\n");
	  block_type[chanNum] = ONLY_SHORT_WINDOW;
	  break;
        case LS_STARTSTOP_SEQUENCE :
	  if (debugLevel>2)
	    printf("WSM: LS_STARTSTOP_SEQUENCE\n");
	  {
	    static enum WINDOW_TYPE long_short_seq[3] = 
            { ONLY_LONG_WINDOW, LONG_SHORT_WINDOW, SHORT_LONG_WINDOW };
	    static int ls_seq_no[MAX_TIME_CHANNELS];
	  
	    block_type[chanNum] = long_short_seq[ls_seq_no[chanNum]++];
	    if( ls_seq_no[chanNum] == 3 ) ls_seq_no[chanNum] = 0;
	  }
	  break;
        case LONG_SHORT_SEQUENCE :
	  if (debugLevel>2)
	    printf("WSM: LONG_SHORT_SEQUENCE\n");
	  {
	    static enum WINDOW_TYPE long_short_seq[4] = 
            { ONLY_LONG_WINDOW, LONG_SHORT_WINDOW, ONLY_SHORT_WINDOW, SHORT_LONG_WINDOW };
	    static int ls_seq_no[MAX_TIME_CHANNELS];

	    block_type[chanNum] = long_short_seq[ls_seq_no[chanNum]++];
	    if( ls_seq_no[chanNum] == 4 ) ls_seq_no[chanNum] = 0;
	  }
	  break;
        case FFT_PE_WINDOW_SWITCHING :
	  {
	  
	    /* A few definitions:                                                      */
	    /*   block_type:  Initially, the block_type used in the previous frame.    */
	    /*                Will be set to the block_type to use this frame.         */
	    /*                A block type will be selected to ensure a meaningful     */
	    /*                window transition.                                       */
	    /*   next_desired_block_type:  Block_type (LONG or SHORT) which the psycho */
	    /*                model wants to use next frame.  The psycho model is      */
	    /*                using a look-ahead buffer.                               */
	    /*   desired_block_type:  Block_type (LONG or SHORT) which the psycho      */
  	    /*                previously wanted to use.  It is the desired block_type  */
	    /*                for this frame.                                          */
	    if ( (block_type[chanNum]==ONLY_SHORT_WINDOW)||(block_type[chanNum]==LONG_SHORT_WINDOW) ) {
	      if ( (desired_block_type[chanNum]==ONLY_LONG_WINDOW)&&(next_desired_block_type[chanNum]==ONLY_LONG_WINDOW) ) {
	        block_type[chanNum]=SHORT_LONG_WINDOW;
	      } else {
	        block_type[chanNum]=ONLY_SHORT_WINDOW;
	      }
	    } else if (next_desired_block_type[chanNum]==ONLY_SHORT_WINDOW) {
	    block_type[chanNum]=LONG_SHORT_WINDOW;
	    } else {
	    block_type[chanNum]=ONLY_LONG_WINDOW;
	    }
	    desired_block_type[chanNum]=next_desired_block_type[chanNum];

	  }
	  break;

        default:
	  CommonExit( 1, "Encode: unsupported window switching mode %d", win_switch_mode );
        }
      }
    }


//block_type[chanNum] = ONLY_SHORT_WINDOW;


    {
      int chanNum;
      /*      for (chanNum=0;chanNum<max_ch;chanNum++) {*/
      for (chanNum=0;chanNum<nchans;chanNum++) {
	switch( block_type[chanNum] ) {
	case ONLY_SHORT_WINDOW  :

	  /*NOTE:  following 3 vars are either dead or NYI.*/
	  sub_win_size = block_size_samples/short_win_in_long;
	  bw_lines     = bw_lines_short;
	  no_sub_win   = short_win_in_long;

#ifdef I2R_LOSSLESS
	  quantInfo[chanNum].max_sfb = chpo_short[chanNum][0].no_of_cb;
#else
		quantInfo[chanNum].max_sfb = 12;           /* Currently not fs dependent (should be) */
#endif
	  quantInfo[chanNum].num_window_groups=4;
	  quantInfo[chanNum].window_group_length[0] = 1;
	  quantInfo[chanNum].window_group_length[1] = 2;
	  quantInfo[chanNum].window_group_length[2] = 3;
	  quantInfo[chanNum].window_group_length[3] = 2;
	  break;
      
	default:

	  /*NOTE:  following 3 vars are either dead or NYI.*/
	  sub_win_size = block_size_samples;
	  bw_lines     = bw_lines_long;
	  no_sub_win   = 1;
	
#ifdef I2R_LOSSLESS
	  	  quantInfo[chanNum].max_sfb = chpo_long[chanNum].no_of_cb;
#else
	
	  if ((sampling_rate == 96000)||(sampling_rate == 192000))
		quantInfo[chanNum].max_sfb = 32; 
	  else
		  quantInfo[chanNum].max_sfb = 40; 
#endif
	  quantInfo[chanNum].num_window_groups = 1;
	  quantInfo[chanNum].window_group_length[0]=1;
	  break;
	} /* end case */
	} /* end for loop */

      /*	  quantInfo[cplidx].max_sfb = 40;
      	  quantInfo[cplidx].num_window_groups = 1;
	  quantInfo[cplidx].window_group_length[0]=1;*/
    }


  /******************************************************************************************************************************
  *
  * pre processing 
  *
  ******************************************************************************************************************************/

#ifdef SON_AAC_PP
    {
    int chanNum;
    for (chanNum=0;chanNum<nchans;chanNum++) {

      /* son_NBCpp */
      if (pp_select == AAC_PP)
	{
	  int band;
	  if(debugLevel>2) printf("pp_aac selected Gain Controled\n");	
	  /* polyphase quadrature filter */
	  /* son_pqf_main(DTimeSigLookAheadBufForGC[chanNum], block_size_samples, MONO_CHAN, DBandSigLookAheadBuf[chanNum]);*/
	  son_pqf_main(DTimeSigLookAheadBufForGC[chanNum], block_size_samples, chanNum, DBandSigLookAheadBuf[chanNum]);

	  /* set analysis buffer for gain control */
	  for (band = 0; band < NBANDS; band++) {
	    memcpy((char *)&DBandSigBufForGCAnalysis[chanNum][band][block_size_samples/NBANDS*2],
		   (char *)DBandSigLookAheadBuf[chanNum][band],
		   block_size_samples/NBANDS*sizeof(double)
		   );

	  }

	if(gc_switch == GC_PRESENT){

	  /* gain detector */
	  son_gc_detect(DBandSigBufForGCAnalysis[chanNum], block_size_samples, block_type[chanNum], chanNum, gainInfo[chanNum]);
	}
	else /* if GC_NON_PRESENT */
	{
	 /* reset detect if no gain control */
     son_gc_detect_reset(DBandSigBufForGCAnalysis[chanNum], block_size_samples, block_type[chanNum], chanNum, gainInfo[chanNum]);
	}


	  /* gain modifier */
	  son_gc_modifier(DBandSigBufForGCAnalysis[chanNum],
			  gainInfo[chanNum],
			  block_size_samples,
			  block_type[chanNum], 
			  chanNum,
			  DBandSigWithGCBuf[chanNum]
			  );

	  /* shift analysis buffer */
	  for (band = 0; band < NBANDS; band++) {
	    memcpy((char *)DBandSigBufForGCAnalysis[chanNum][band],
		   (char *)&DBandSigBufForGCAnalysis[chanNum][band][block_size_samples/NBANDS],
		   block_size_samples/NBANDS*2*sizeof(double)
		   );
	  }

		if(gc_switch == GC_PRESENT){
	  	/* packing gain data into the gc_stream */
	  	son_gc_pack(gc_stream[chanNum], block_type[chanNum],
		      max_band[chanNum], gainInfo[chanNum]);
		}
	}

      /* Count number of bits used for gain_control_data */
      /*      used_bits += WriteGainControlData(&quantInfo[chanNum], */    /* quantInfo contains packed gain control data */
      WriteGainControlData(&quantInfo[chanNum],     /* quantInfo contains packed gain control data */
					NULL,           /* NULL BsBitStream.  Only counting bits, no need to write yet */
					0);             /* Zero write flag means don't write */
    }
    }
#else   /* Gain control calls bypassed */
      BsPutBit(gc_stream[chanNum],0,LEN_GAIN_PRES); /* switches off the gain control */
      /*used_bits += LEN_GAIN_PRES;*/
      /*printf("PP_AAC not defined \n");	*/
#endif
  }
    


  /******************************************************************************************************************************
  *
  * T/F mapping
  *
  ******************************************************************************************************************************/

  {
    /* son_NBCpp */
    if (pp_select == AAC_PP)
      {
	int i, band;
	double tmp;
	int chanNum;
	for (chanNum=0;chanNum<nchans;chanNum++) {
	  for (band = 0; band < NBANDS; band++) {
	    buffer2freq(
			DBandSigWithGCBuf[chanNum][band],
			&spectral_line_vector_for_gc[chanNum][block_size_samples/NBANDS*band],
			overlap_buffer[chanNum],
			block_type[chanNum],
			WS_FHG ,
			block_size_samples/NBANDS, 
			block_size_samples/NBANDS/med_win_in_long,
			block_size_samples/NBANDS/short_win_in_long,	 
			NON_OVERLAPPED
			);

	  } /* for(band=...) */


	  son_gc_arrangeSpecEnc(spectral_line_vector_for_gc[chanNum],
				block_size_samples,block_type[chanNum],
				spectral_line_vector[chanNum]
				);
	}
      } /* if (pp_select == AAC_PP) */
    else
      {
	int chanNum;
	int i;
	for (chanNum=0;chanNum<nchans;chanNum++) {
#ifndef I2R_LOSSLESS
		buffer2freq(
		      DTimeSigBuf[chanNum],
		      spectral_line_vector[chanNum],
		      overlap_buffer[chanNum],
		      block_type[chanNum],
		      WS_FHG,
		      block_size_samples,
		      block_size_samples/med_win_in_long,
		      block_size_samples/short_win_in_long,
		      OVERLAPPED
		      );
#else

		for (i=0;i<osf*1024;i++) {
		  int_DTimeSigBuf[chanNum][i] = (int)( DTimeSigBuf[chanNum][i]/pcm_scaling[ll_info[chanNum].type_PCM]);
		  int_DTimeSigLookAheadBuf[chanNum][i] = (int)( DTimeSigLookAheadBuf[chanNum][i]/pcm_scaling[ll_info[chanNum].type_PCM]);
		}

	}

	if (msStereoIntMDCTMode == 0) {
    msStereoIntMDCT = 0;
	} else {
    msStereoIntMDCT = 1;
	}

  msenable = msStereoIntMDCT;
	//block_type[chanNum] = ONLY_SHORT_WINDOW;

	for (chanNum=0;chanNum<nchans;chanNum++) {
	  if (coreenable) {
	    block_type_current = block_type[chanNum];
	  } else {
	    block_type_current = block_type_lossless_only;
	  }
	  if (channelInfo[chanNum].cpe && msStereoIntMDCT) {
	    /* apply StereoIntMDCT once per CPE */
	    if (channelInfo[chanNum].ch_is_left) {
	      int leftChan=chanNum;
	      int rightChan=channelInfo[chanNum].paired_ch;

	      StereoIntMDCT(int_DTimeSigBuf[leftChan],
			    int_DTimeSigBuf[rightChan],
			    mdctTimeBuf[leftChan],
			    mdctTimeBuf[rightChan],
			    int_spectral_line_vector[leftChan],
			    int_spectral_line_vector[rightChan],
			    block_type_current,
			    WS_FHG,
			    osf);

        /* for IntTNS: */
        {
          /* do inverse M/S to provide L/R spectrum for TNS */
          for (i=0; i<osf*1024; i++) {
            InverseIntMidSideINT32(&int_spectral_line_vector[leftChan][i],
                                   &int_spectral_line_vector[rightChan][i]);
          }
        }

	    }
	  } else {
	    IntMDCT(int_DTimeSigBuf[chanNum],
		    mdctTimeBuf[chanNum],
		    int_spectral_line_vector[chanNum],
		    block_type_current,
		    WS_FHG,
		    osf);
	  }
	}

	if (coreenable) {
	  for (chanNum=0;chanNum<nchans;chanNum++) {
	    for (i=0;i<1024;i++)
	      spectral_line_vector[chanNum][i] = int_spectral_line_vector[chanNum][i] * MDCT_SCALING * pcm_scaling[ll_info[chanNum].type_PCM] * osf_scaling[osf];
	  }
	}
#endif
      }
  }


  /******************************************************************************************************************************
  *
  * adapt ratios of psychoacoustic module to codec scale factor bands
  *
  ******************************************************************************************************************************/

#ifdef I2R_LOSSLESS
   if (coreenable)
   {
#endif

    {
      int chanNum;
      for (chanNum=0;chanNum<nchans;chanNum++) {

	/* Save p_ratio from psychoacoustic model for next frame.  */
	/* Psycho model is using a look-ahead window for block switching */

	memcpy( (char*)p_ratio_long[!psychNum][chanNum], (char*)chpo_long[chanNum].p_ratio, (NSFB_LONG)*sizeof(double) );
	memcpy( (char*)p_ratio_short[!psychNum][chanNum],(char*)chpo_short[chanNum][0].p_ratio,(MAX_SHORT_WINDOWS*NSFB_SHORT)*sizeof(double) );
      }
    }
#ifdef I2R_LOSSLESS
   }
#endif

  {
    int    sfb, i, sl, par_set;
    double energy, dtmp;
    int chanNum;
    for (chanNum=0;chanNum<nchans;chanNum++) {
      par_set = 0;
#ifdef I2R_LOSSLESS
   if (coreenable)
   {
#endif

      switch( block_type[chanNum] ) {
      case ONLY_LONG_WINDOW :
	memcpy( (char*)sfb_width_table[chanNum], (char*)chpo_long[chanNum].cb_width, (NSFB_LONG+1)*sizeof(int) );
	nr_of_sfb[chanNum] = chpo_long[chanNum].no_of_cb;
	p_ratio[chanNum]   = p_ratio_long[psychNum][chanNum];
	par_set = 1;
	break;
      case LONG_SHORT_WINDOW :
	memcpy( (char*)sfb_width_table[chanNum], (char*)chpo_long[chanNum].cb_width, (NSFB_LONG+1)*sizeof(int) );
	nr_of_sfb[chanNum] = chpo_long[chanNum].no_of_cb;
	p_ratio[chanNum]   = p_ratio_long[psychNum][chanNum];
	par_set = 1;
	break;
      case ONLY_SHORT_WINDOW :
	memcpy( (char*)sfb_width_table[chanNum], (char*)chpo_short[chanNum][0].cb_width, (NSFB_SHORT+1)*sizeof(int) );
	nr_of_sfb[chanNum] = chpo_short[chanNum][0].no_of_cb;
	p_ratio[chanNum]   = p_ratio_short[psychNum][chanNum];
	par_set = 1;
	break;
      case SHORT_LONG_WINDOW :
	memcpy( (char*)sfb_width_table[chanNum], (char*)chpo_long[chanNum].cb_width, (NSFB_LONG+1)*sizeof(int) );
	nr_of_sfb[chanNum] = chpo_long[chanNum].no_of_cb;
	p_ratio[chanNum]   = p_ratio_long[psychNum][chanNum];
	par_set = 1;
	break;
      default:
	CommonWarning( "Encode: unsupported block type in adapt ratios: %d", block_type[chanNum] );
	break;
      }
#ifdef I2R_LOSSLESS
   }
#endif

      /* to prevent undefined ptr just use a meaningful value to allow for experiments with the meachanism above */
      if( !par_set) {   
	memcpy( (char*)sfb_width_table[chanNum], (char*)chpo_long[chanNum].cb_width, (NSFB_LONG+1)*sizeof(int) );
	nr_of_sfb[chanNum]  = chpo_long[chanNum].no_of_cb;
	p_ratio[chanNum]   = p_ratio_long[psychNum][chanNum];
      }
#ifdef I2R_LOSSLESS
	  quantInfo[chanNum].total_sfb = /*quantInfo[chanNum].num_window_groups **/ nr_of_sfb[chanNum];
#endif

    }
  }

    {
      int chanNum;   
      for (chanNum=0;chanNum<nchans;chanNum++) {
	/* Construct sf band offset table */
	int offset=0;
	int sfb;
	for (sfb=0;sfb<nr_of_sfb[chanNum];sfb++) {
	  sfb_offset_table[chanNum][sfb] = offset;
	  offset+=sfb_width_table[chanNum][sfb];
	}
	sfb_offset_table[chanNum][nr_of_sfb[chanNum]]=offset;
      }
    }



  /************************************************************************
   Set upper spectral coefficients to zero for LFE
   ************************************************************************/
  {
    int chanNum;
    for (chanNum=0;chanNum<max_ch;chanNum++) { 
      if (channelInfo[chanNum].lfe) {           
        int i;    
        for (i=13;i<sfb_offset_table[chanNum][nr_of_sfb[chanNum]];i++)      
		{
          spectral_line_vector[chanNum][i] = 0;
#ifdef I2R_LOSSLESS
		  int_spectral_line_vector[chanNum][i] = 0;
#endif
		}

		
      }
    }               
  }


  /******************************************************************************************************************************
  *
  * quantization and coding
  *
  ******************************************************************************************************************************/
  { 
    int padding_limit = max_bitreservoir_bits;    /* f( max__bitreservoir_bits ) */
    int maxNumBitsByteAligned;
    int chanNum;   
    int numWindows;
    int windowLength;
    int j,w;
    int bandNumber;
    int numFillBits;
    int bitsLeftAfterFill;
	int cpl_bits;

    /* bit budget */
    num_bits_available = (long)(average_bits + available_bitreservoir_bits - used_bits);
    
    /* find the largest byte-aligned section with fewer bits than num_bits_available */
    maxNumBitsByteAligned = ((num_bits_available >> 3) << 3);

    /* Compute how many reservoir bits can be used and still be able to byte */
    /* align without exceeding num_bits_available, and have room for an ID_END marker   */
    available_bitreservoir_bits = maxNumBitsByteAligned - LEN_SE_ID - average_bits;
#if 0
    for (chanNum=0;chanNum<nchans;chanNum++) {

	  /* Set window shape paremeter in quantInfo */
      if (channelInfo[chanNum].lfe) {
        quantInfo[chanNum].window_shape = WS_FHG;   /* for LFE window_shape must be sine */
      } else {
        quantInfo[chanNum].window_shape = WS_FHG;   /* other window_shapes can be selected */
      }

      /* Compute energy in each scalefactor band of each window */
      numWindows = (block_type[chanNum]==ONLY_SHORT_WINDOW) ?
	short_win_in_long : 1;
      windowLength = block_size_samples/numWindows;
      bandNumber=0;
      for (w=0;w<numWindows;w++) {
	int offset=0;
	int sfb;
	j = w*windowLength;

	/* Only compute energy up to max_sfb */
	for(sfb=0; sfb< quantInfo[chanNum].max_sfb; sfb++ ) {
	  /* calculate scale factor band energy */
	  int width,i;
	  energy[chanNum][bandNumber] = 0.0;
	  width=sfb_width_table[chanNum][sfb];
	  for(i=offset; i<(offset+width); i++ ) {
#ifdef I2R_LOSSLESS
	   double dtmp = (double)(int_spectral_line_vector[chanNum][j++]*MDCT_SCALING);
#else
	    double dtmp = spectral_line_vector[chanNum][j++];
#endif
	    energy[chanNum][bandNumber] += dtmp*dtmp;
	  }
	  bandNumber++;
	  offset+=width;
	}  
      }
    }  /* for (chanNum... */
#endif
    /******************************************/
    /* Perform Coupling Coordinate calculation*/
    /******************************************/
 
	if (cplenabl)
	{
		couple_init(&cvars, nfbwchans, quantInfo, channelInfo);
	}

    /******************************************/
    /* Perform TNS analysis and filtering     */
    /******************************************/
    for (chanNum=0;chanNum<nchans;chanNum++) {
      if ((!channelInfo[chanNum].lfe)&&(tnsenable)) {
        TnsEncode(nr_of_sfb[chanNum],            /* Number of bands per window */
	          quantInfo[chanNum].max_sfb,              /* max_sfb */
                  block_type[chanNum],
                  sfb_offset_table[chanNum],
                  spectral_line_vector[chanNum],
                  &tnsInfo[chanNum]);
        
        /* IntTNS */
        {
          IntTnsEncodeFilterOnly(nr_of_sfb[chanNum],            /* Number of bands per window */
                                 quantInfo[chanNum].max_sfb,              /* max_sfb */
                                 block_type[chanNum],
                                 sfb_offset_table[chanNum],
                                 int_spectral_line_vector[chanNum],
                                 &tnsInfo[chanNum]);
        }
      } else {
        tnsInfo[chanNum].tnsDataPresent=0;      /* TNS not used for LFE */
      }
    }

    /* for IntTNS: */
    {
      for (chanNum=0;chanNum<nchans;chanNum++) {
        if (channelInfo[chanNum].cpe && msStereoIntMDCT && 
            channelInfo[chanNum].ch_is_left) {
          int leftChan=chanNum;
          int rightChan=channelInfo[chanNum].paired_ch;
          int i;
          
          /* do forward M/S to revert to M/S spectrum after TNS */
          for (i=0; i<osf*1024; i++) {
            IntMidSideINT32(&int_spectral_line_vector[leftChan][i],
                            &int_spectral_line_vector[rightChan][i]);
          }
        } 
      }
    }

    /*******************************************************************************/
    /* If LTP prediction is used, compute LTP predictor info and residual spectrum */
    /*******************************************************************************/
#ifdef MPEG4V1
    {
      TNS_INFO *tnsInfo_for_LTP;

      for(chanNum=0;chanNum<max_ch;chanNum++)
      {
	if((tnsInfo[chanNum].tnsDataPresent != 0) && tnsenable)
	  tnsInfo_for_LTP = &tnsInfo[chanNum];
	else
	  tnsInfo_for_LTP = NULL;
#ifndef I2R_LOSSLESS /*disable LTP in this lossless RM encoder*/	  
	if(channelInfo[chanNum].present && (!channelInfo[chanNum].lfe) &&
	   block_type[chanNum] != ONLY_SHORT_WINDOW) 
	{
	  nok_ltp_enc(spectral_line_vector[chanNum], 
		      nok_tmp_DTimeSigBuf[chanNum], 
		      block_type[chanNum],
		      WS_FHG, WS_FHG,
		      block_size_samples,
		      block_size_samples/med_win_in_long,
		      block_size_samples/short_win_in_long, 
		      &sfb_offset_table[chanNum][0], 
		      nr_of_sfb[chanNum],
		      &nok_lt_status[chanNum],
		      tnsInfo_for_LTP);
	} /* if(channelInfo[chanNum].present... */
	else
#endif
	  quantInfo[chanNum].ltpInfo->global_pred_flag = 0;
	
      } /* for(chanNum... */
    }
#endif

#ifndef I2R_LOSSLESS /* disable IS in this lossless RM encoder */
    /******************************************/
    /* Apply Intensity Stereo                 */
    /******************************************/
#ifdef MPEG4V1
    if ( pns_sfb_start > PNS_MAX_START )  /* do intensity only if pns is off  */
#endif
      {
        ISEncode(spectral_line_vector,
                 channelInfo,
                 sfb_offset_table,
                 block_type,
                 quantInfo,
                 max_ch,
		 cplenabl,
		 isenable);
      }

    /***********************************************************************/
    /* If prediction is used, compute predictor info and residual spectrum */
    /***********************************************************************/
#ifndef MPEG4V1
    for (chanNum=0;chanNum<nchans;chanNum++) {      
      if ((qc_select == AAC_PRED)&&(!channelInfo[chanNum].lfe)) {  /* no prediction for LFE channel*/      
	int numPredBands = min(max_pred_sfb,nr_of_sfb[chanNum]);
	PredCalcPrediction( spectral_line_vector[chanNum], 
			   reconstructed_spectrum[chanNum], 
			   (int)block_type[chanNum], 
			   numPredBands,
			   sfb_width_table[chanNum], 
			   /* Added by CL */
			   quantInfo,
			   channelInfo,
			   chanNum); 
      } else {
	quantInfo[chanNum].pred_global_flag = 0;
      }
    } /* for(chanNum... */
#endif
#endif /* I2R_LOSSLESS */
    /******************************************/
    /* Apply MS stereo                        */
    /******************************************/
#ifndef I2R_LOSSLESS
    MSEncode(spectral_line_vector,
	     channelInfo,
	     sfb_offset_table,
	     block_type,
	     quantInfo,
	     max_ch,
	     msenable);
#else  

    /* integer M/S */
    lit_MSEncode(int_spectral_line_vector,
                 channelInfo,
                 sfb_offset_table,
                 block_type,
                 quantInfo,
                 max_ch,
                 msenable);

    /* prepare for AAC_QC */

		for (chanNum = 0;chanNum<nchans;chanNum++)
		{
			int i;
		    if (channelInfo[chanNum].present) 
			{
				  if ((channelInfo[chanNum].cpe)&&(channelInfo[chanNum].ch_is_left)) 
				  {
					int leftChan=chanNum;
					int rightChan=channelInfo[chanNum].paired_ch;
          
         /*  channelInfo[leftChan].common_window = 1; */
          

					if (channelInfo[leftChan].ms_info.is_present == 2) /* only "all on" supported now*/
					{
						ll_info[leftChan].CoreScaling =  \
										(block_type[leftChan] == ONLY_SHORT_WINDOW)?MDCT_SCALING_SHORT_MS:MDCT_SCALING_MS;
						ll_info[rightChan].CoreScaling = \
										(block_type[rightChan] == ONLY_SHORT_WINDOW)?MDCT_SCALING_SHORT_MS:MDCT_SCALING_MS;
					}
					else
					{
						ll_info[leftChan].CoreScaling =  \
							(block_type[leftChan] == ONLY_SHORT_WINDOW)?MDCT_SCALING_SHORT:MDCT_SCALING;
						ll_info[rightChan].CoreScaling = \
							(block_type[rightChan] == ONLY_SHORT_WINDOW)?MDCT_SCALING_SHORT:MDCT_SCALING;
					}

				  }
				  else if (!channelInfo[chanNum].cpe)
					  ll_info[chanNum].CoreScaling = (block_type[chanNum] == ONLY_SHORT_WINDOW)?MDCT_SCALING_SHORT:MDCT_SCALING;
			}
			ll_info[chanNum].CoreScaling *= 
        pcm_scaling[ll_info[chanNum].type_PCM] * osf_scaling[osf];

	  		for (i=0;i<1024;i++)
			{
 				spectral_line_vector[chanNum][i] = (double)(int_spectral_line_vector[chanNum][i]/* +(double)rand()*6/RAND_MAX-3 */)  *\
					ll_info[chanNum].CoreScaling;
			}

		}

#endif

#if 1
    for (chanNum=0;chanNum<nchans;chanNum++) {

	  /* Set window shape paremeter in quantInfo */
      if (channelInfo[chanNum].lfe) {
        quantInfo[chanNum].window_shape = WS_FHG;   /* for LFE window_shape must be sine */
      } else {
        quantInfo[chanNum].window_shape = WS_FHG;   /* other window_shapes can be selected */
      }

      /* Compute energy in each scalefactor band of each window */
      numWindows = (block_type[chanNum]==ONLY_SHORT_WINDOW) ?
					short_win_in_long : 1;
      windowLength = block_size_samples/numWindows;
      bandNumber=0;
      for (w=0;w<numWindows;w++) {
	int offset=0;
	int sfb;
	j = w*windowLength;

	/* Only compute energy up to max_sfb */
	for(sfb=0; sfb< quantInfo[chanNum].total_sfb; sfb++ ) {
	  /* calculate scale factor band energy */
	  int width,i;
	  energy[chanNum][bandNumber] = 0.0;
	  width=sfb_width_table[chanNum][sfb];
	  for(i=offset; i<(offset+width); i++ ) {
		  double dtmp = spectral_line_vector[chanNum][j++];
		  energy[chanNum][bandNumber] += dtmp*dtmp;
	  }
	  bandNumber++;
	  offset+=width;
	}  
      }
    }  /* for (chanNum... */
#endif




   /************************************************/
   /* Set the dynamic range parameters             */
   /************************************************/


	if (dvars_inpt->dynrng_enabl)
	{
		drcbits = setDynRngVars(dvars_inpt, block_type, nchans);
	}
	else
	{
		drcbits = 0;
	}

   /************************************************/
   /* Call the AAC quantization and coding module. */
   /************************************************/


#ifdef BSAC
	if (codec_type == 1)
	available_bitreservoir_bits = 0;//SLS_BSAC
#endif
   for (chanNum=0;chanNum<nchans;chanNum++) {

     int bitsToUse; 
     if (!lfePresent) {
       bitsToUse = (int)((average_bits - used_bits)/nchans);
       bitsToUse += (int)(0.2*available_bitreservoir_bits/nchans);
     } else {
       /* Calculation of reduced bitrate for LFE */
       double lfeBitRatio = 0.14;       /* ratio of LFE bits to bits of one SCE */
       int lfeBits = max(200,(int)((average_bits - used_bits) * lfeBitRatio / (nchans - 1))); /* number of bits for LFE */ 
       
       if (channelInfo[chanNum].lfe) {
         bitsToUse = lfeBits;
         bitsToUse += (int)(0.2*available_bitreservoir_bits 
                            * lfeBitRatio / (nchans - 1));
       } else {
         bitsToUse = (int)((average_bits - used_bits - lfeBits) / (nchans - 1));
         bitsToUse += (int)(0.2*available_bitreservoir_bits 
                            * (1 - lfeBitRatio / (nchans - 1)) / (nchans - 1));
       } /* if (channelInfo[chanNum].lfe) */
     } /* if (!lfePresent) */  

	if (chanNum == max_ch)
	{
		cpl_bits = calc_cplbits(nfbwchans, cplmix, cplenabl, channelInfo, &cvars);
		bitsToUse -= cpl_bits;
	}

	if (chanNum == 0)	   /* subtract off bits used for dynamic range control */
	{
		bitsToUse -= drcbits;		
	}


#ifdef I2R_LOSSLESS
   if (!coreenable)
   {
	   int i;
	   for (i=0;i<1024;i++)
		   spectral_line_vector[chanNum][i] = 0;
	   //printf("core disabled\n");
   }
	   
#endif

     tf_encode_spectrum_aac( &spectral_line_vector[chanNum],
			     &p_ratio[chanNum],
			     &allowed_distortion[chanNum],
			     &energy[chanNum],
			     &block_type[chanNum],
			     &sfb_width_table[chanNum],
			     &nr_of_sfb[chanNum],
			     bitsToUse,
			     padding_limit,
			     fixed_stream, /*tmp_stream,*/
			     var_stream,
			     1,                        /* nr of audio channels */
			     &reconstructed_spectrum[chanNum],
			     useShortWindows,
			     aacAllowScalefacs,
			     &quantInfo[chanNum],
#ifdef I2R_LOSSLESS
				 &quant_aac[chanNum],
#endif
			     &(channelInfo[chanNum])
				 );


#ifdef I2R_LOSSLESS
 
		if (block_type[chanNum] == ONLY_SHORT_WINDOW)
		 {
		   lit_ll_sfg(&quantInfo[chanNum],
			      sfb_width_table[chanNum],
			      int_spectral_line_vector[chanNum],
			      osf
			      );
		 }
   }
   	for (chanNum=0;chanNum<nchans;chanNum++) {
		if ((channelInfo[chanNum].cpe)&&(channelInfo[chanNum].ch_is_left)) {
			int leftChan=chanNum;
			int rightChan=channelInfo[chanNum].paired_ch;

		 LIT_ll_errorMap(
				int_spectral_line_vector[leftChan],  /* MDCT spectrum */
				&quantInfo[leftChan],      /* AAC quantization information */ 
				&ll_info[leftChan],
				quant_aac[leftChan],
				&channelInfo[leftChan],
				msenable,
				osf,
				block_type_lossless_only,
				chpo_short[0][0].cb_width,
				chpo_short[0][0].no_of_cb,
				coreenable
				);			
		 		
		 LIT_ll_errorMap(
				int_spectral_line_vector[rightChan],  /* MDCT spectrum */
				&quantInfo[rightChan],      /* AAC quantization information */ 
				&ll_info[rightChan],
				quant_aac[rightChan],
				&channelInfo[leftChan],
				msenable,
				osf,
				block_type_lossless_only,
				chpo_short[0][0].cb_width,
				chpo_short[0][0].no_of_cb,
				coreenable
				);			

		}
		else if (!channelInfo[chanNum].cpe)
		{
		 LIT_ll_errorMap(
				int_spectral_line_vector[chanNum],  /* MDCT spectrum */
				&quantInfo[chanNum],      /* AAC quantization information */ 
				&ll_info[chanNum],
				quant_aac[chanNum],
				&channelInfo[chanNum],
				msenable,
				osf,
				block_type_lossless_only,
				chpo_short[0][0].cb_width,
				chpo_short[0][0].no_of_cb,
				coreenable
				);			

		}
	}
	   for (chanNum=0;chanNum<nchans;chanNum++) {
		 LIT_ll_perceptualEnhEnc(
			 &ll_info[chanNum],
			 &quantInfo[chanNum],
			 ll_bits_to_used/nchans,
			 sampling_rate*osf,		// I2R: modified to osf*core_sample_rate (28 Jun 07)
			 osf,
			 chpo_short[0][0].no_of_cb,
			 coreenable,
			 &channelInfo[chanNum],
			 block_type_lossless_only,
			 msStereoIntMDCT
			 );



#endif

   }




   
   /**********************************************************/
   /* Reconstruct MS Stereo bands for prediction             */
   /**********************************************************/
#if 0
#ifndef I2R_LOSSLESS
   MSReconstruct(reconstructed_spectrum,
		 channelInfo,
		 sfb_offset_table,
		 block_type,
		 quantInfo,
		 max_ch);

#else
     lit_MSReconstruct(rec_spectrum,
		 channelInfo,
		 sfb_offset_table,
		 block_type,
		 quantInfo,
		 max_ch);

	for (chanNum=0;chanNum<nchans;chanNum++) {      
			int i;
			for (i=0;i<1024;i++)
				reconstructed_spectrum[chanNum][i] = (double)rec_spectrum[chanNum][i] *MDCT_SCALING;
		}


#endif
#endif	 
   /**********************************************************/
    /* Reconstruct Intensity Stereo bands for prediction     */
   /**********************************************************/
#ifndef I2R_LOSSLESS
	 ISReconstruct(reconstructed_spectrum,
		 channelInfo,
		 sfb_offset_table,
		 block_type,
		 quantInfo,
		 max_ch);
 
#endif

   /**********************************************************/
   /* Add the predicted spectrum to the so far recontructed  */
   /* spectrum, then TNS decode (if needed) and finally      */
   /* update LTP history buffer                              */
   /**********************************************************/
#if 0 /* turn off for this LOSSLESS RM */
#ifdef MPEG4V1
   {
	   
     TNS_INFO *tnsDecInfo;

     for (chanNum=0;chanNum<max_ch;chanNum++) 
     {
       if((tnsInfo[chanNum].tnsDataPresent != 0) && tnsenable)
	 tnsDecInfo = &tnsInfo[chanNum];
       else
	 tnsDecInfo = NULL;
       
       if (!channelInfo[chanNum].lfe) {  /* no reconstruction needed for LFE channel*/

	 nok_ltp_reconstruct(reconstructed_spectrum[chanNum], 
			     block_type[chanNum],
			     &sfb_offset_table[chanNum][0], 
			     nr_of_sfb[chanNum],
			     &nok_lt_status[chanNum]);
	 if(tnsDecInfo != NULL)
	   TnsDecodeFilterOnly(nr_of_sfb[chanNum], quantInfo[chanNum].max_sfb, 
			       block_type[chanNum], &sfb_offset_table[chanNum][0], 
			       reconstructed_spectrum[chanNum], tnsDecInfo);

	 
	 freq2buffer(reconstructed_spectrum[chanNum], time_buffer[chanNum], 
		     ltp_overlap_buffer[chanNum], block_type[chanNum], 
		     block_size_samples, block_size_samples/med_win_in_long,
		     block_size_samples/short_win_in_long, WS_FHG, OVERLAPPED);

	 nok_ltp_update(&nok_lt_status[chanNum], time_buffer[chanNum],
			ltp_overlap_buffer[chanNum], block_size_samples);

#ifdef I2R_LOSSLESS /*local decoder for debuging */
	 {
		int i;
		
	/*	freq2buffer(reconstructed_spectrum[chanNum], lossless_time_buffer[chanNum], 
		     lossless_overlap_buffer[chanNum], block_type[chanNum], 
		     block_size_samples, block_size_samples/med_win_in_long,
		     block_size_samples/short_win_in_long, WS_FHG, OVERLAPPED);
      */  
		lit_ll_freq2buffer(rec_spectrum[chanNum],mdctFreqBuf[chanNum],lossless_time_buffer[chanNum]);
 		for (i=0;i<1024;i++)
		{
			 lossless_time_error[chanNum][i] = (int)p_prev2_time_signal[chanNum][i] - lossless_time_buffer[chanNum][i]; /* rounding */
			 p_prev2_time_signal[chanNum][i] = p_prev_time_signal[chanNum][i];
			 p_prev_time_signal[chanNum][i] = p_time_signal[chanNum][i];
		}
	 }
 
	 
#endif /* lossless*/

       }
     }
   }
#endif

#endif // if 0

   /* If short window, reconstruction not needed for prediction */
   for (chanNum=0;chanNum<max_ch;chanNum++) {
     if ((block_type[chanNum]==ONLY_SHORT_WINDOW)) {
       int sind;
       for (sind=0;sind<1024;sind++) {
	 reconstructed_spectrum[chanNum][sind]=0.0;
       }
     }
   }

#ifdef BSAC
   //SLS_BSAC
   if (codec_type == 1) 
  	{
	   int MapAacBsac[6] = {1,2,0,5,3,4};


		/*typedef struct {
		  int ms_mask;
		  int ms_used[MAX_SHORT_WINDOWS][SFB_NUM_MAX];
		} MSInfo;*/
		/*struct MSInfo msInfo;
		msInfo.ms_used = 2;
		for( ch=0; ch<nr_of_chan-1; ch++ )
		for (sb=0; sb<quantInfo -> total_sfb; sb++)
			msInfo.ms_mask[ch][sb] = 1;*/
	   MS_Info *ms[10];
	   TNS_INFO *tns[10];
		int ch,sb;
		int scfacs[6][17*8];
		int quant[6][1024];
	   int i_elm;
	   int chNum;
	   int startChIdx;
	   int endChIdx;
	   int ch_type;
	   int bits_written = 0;
	   int mc_enabled = 0;
	   int numElm = bsac_ex_info->ch_elm_total;
	   enum WINDOW_TYPE window_shape[MAX_TIME_CHANNELS];
	   int num_window_groups[MAX_TIME_CHANNELS];
	   int window_group_length[MAX_TIME_CHANNELS][MAX_SHORT_IN_LONG_BLOCK];
	   int ext_flag = 1;
	   if (max_ch == 2)
	   {
		   MapAacBsac[0] = 0;
		   MapAacBsac[1] = 1;
	   }
	   for( ch=0; ch<max_ch; ch++ )
		{
		   int gp;
		   for (sb=0; sb<quantInfo[ch].total_sfb; sb++)
				scfacs[ch][sb] = quantInfo[ch].scale_factor[sb];
			for (sb=0; sb<1024; sb++)
				quant[ch][sb] = quant_aac[ch][sb];
			tns[ch] = &tnsInfo[ch];
			ms[ch] = &channelInfo[ch].ms_info;
		   num_window_groups[ch] = quantInfo[ch].num_window_groups;
		   for (gp = 0 ; gp<num_window_groups[ch] ; gp++)
			   window_group_length[ch][gp] = quantInfo[ch].window_group_length[gp];
		}
	   if(numElm>1) mc_enabled = 1;
		

	   for(i_elm = numElm-1; i_elm>=0; i_elm--)
	   {
		   HANDLE_CHANNEL_ELEMENT_DATA pChData = &bsac_ex_info->chElmData[i_elm];
		   chNum = pChData->chNum;
		   startChIdx = MapAacBsac[pChData->startChIdx];
		   endChIdx = MapAacBsac[pChData->endChIdx];
		   ch_type =  pChData->ch_type;

		   bits_written += sam_encode_frame(
			   chNum, /*num_of_chan*/
			   ch_type, /*ch_type*/
			   block_type[startChIdx], /*windowSequence*/
			   quantInfo[startChIdx].window_shape,/*window_shape*/						
			   sfb_offset_table[startChIdx], /*sfb_offset[]*/
			   nr_of_sfb[startChIdx]*num_window_groups[startChIdx],/*nr_of_sfs*/
			   num_window_groups[startChIdx], /*num_window_groups*/
			   &window_group_length[startChIdx]/*window_group_length[]*/,
			   &quant[startChIdx], /*quant[][1024]*/
			   &scfacs[startChIdx], /*scfacs[][MAX_SCFAC_BANDS]*/
			   NULL, /**msInfo*/
			   fixed_stream, /*fixed_stream*/
			   0,/*BSAC_SLS i_ch*/ 
			   0, /*avr_bits*/
			   pChData->extension ? 0 : bits_written,/*mc_ext_size*/
			   mc_enabled, /*mc_enabled*/
			   pChData->extension/*extension_flag*/,
			   &tns[startChIdx],/*tns_info[]*/
			   &ms[startChIdx]/*ms_info[]*/
			   );
	   }
	}
#endif



   /**********************************/
   /* Write out all encoded channels */
   /**********************************/
   { 
     for (chanNum=0;chanNum<max_ch;chanNum++) {
       if (channelInfo[chanNum].present) {
	 /* Write out a single_channel_element */
	 if (!channelInfo[chanNum].cpe) {
           if (channelInfo[chanNum].lfe) {
#ifdef BSAC
			   if (codec_type == 0)
#endif
    	     /* Write out lfe */ 
	     used_bits += WriteLFE(&quantInfo[chanNum],   /* Quantization information */
				  channelInfo[chanNum].tag,
				  fixed_stream,           /* Bitstream */
				  1);                     /* Write flag, 1 means write */
           } else {
#ifdef BSAC
			   if (codec_type == 0)
#endif
             /* Write out sce */ /* BugFix by YT  '+=' sould be '=' */
	     used_bits += WriteSCE(&quantInfo[chanNum],   /* Quantization information */
				  channelInfo[chanNum].tag,
				  fixed_stream,           /* Bitstream */
				  1);                     /* Write flag, 1 means write */
#ifdef I2R_LOSSLESS
		 Write_ll_SCE(&ll_info[chanNum],&quantInfo[chanNum],ll_stream,1);
#endif
           } /* if (channelInfo[chanNum].lfe ... else */
	 } else {
	   if (channelInfo[chanNum].ch_is_left) {
	     /* Write out cpe */
#ifdef BSAC
		   if (codec_type == 0)
#endif
	     used_bits += WriteCPE(&quantInfo[chanNum],   /* Quantization information,left */
				  &quantInfo[channelInfo[chanNum].paired_ch],   /* Right */
				  channelInfo[chanNum].tag,
				  channelInfo[chanNum].common_window,    /* common window */
				  &(channelInfo[chanNum].ms_info),
				  fixed_stream,           /* Bitstream */
				  1);                     /* Write flag, 1 means write */
#ifdef I2R_LOSSLESS
		 Write_ll_SCE(&ll_info[chanNum],&quantInfo[chanNum],ll_stream,1);
		 Write_ll_SCE(&ll_info[channelInfo[chanNum].paired_ch],&quantInfo[channelInfo[chanNum].paired_ch],ll_stream,1);
#endif
	   }
	 }  /* if (!channelInfo[chanNum].cpe)  else */
       } /* if (chann...*/
     } /* for (chanNum...*/
   }

#ifdef BSAC
   if (codec_type == 0)
   {
#endif
	if (cplenabl)
	{
		used_bits += WriteCplData (fixed_stream,		/* in: bit stream */
								   &cvars,				/* coupling params */
								   &quantInfo[max_ch],	/* quantization params */
								   max_ch);				/* number of channels */
	}

	if (dvars_inpt->dynrng_enabl)
	{
		used_bits += WriteDRCArray (fixed_stream, nchans);
	}

	/* Compute how many fill bits are needed to avoid overflowing bit reservoir */
   /* Save room for ID_END terminator */
   if (used_bits < (minBitsNoOverflow - LEN_SE_ID) ) {
     numFillBits = minBitsNoOverflow - LEN_SE_ID - used_bits;
   } else {
     numFillBits = 0;
   }
   
   /* Write AAC fill_element, smallest fill element is 7 bits. */
   /* Function may leave up to 6 bits left after fill, so tell it to fill a few extra */

   numFillBits += 6;
   bitsLeftAfterFill=WriteAACFillBits(fixed_stream,numFillBits);
   used_bits += (numFillBits - bitsLeftAfterFill);

   /* Write ID_END terminator */
   BsPutBit(fixed_stream,ID_END,LEN_SE_ID);
   used_bits += LEN_SE_ID;
   
   /* Now byte align the bitstream */
   used_bits += ByteAlign(fixed_stream);
#ifdef BSAC
   }
#endif

   BsClose(fixed_stream);
#ifdef I2R_LOSSLESS
   BsClose(ll_stream);
#endif
   if (var_stream != NULL)
     BsClose(var_stream);
   BsClose(tmp_stream);
   BsFreeBuffer(tmpBitBuf);
   for (chanNum=0;chanNum<MAX_TIME_CHANNELS;chanNum++) {
     BsClose(gc_stream[chanNum]);
     BsFreeBuffer(quantInfo[chanNum].gcBitBuf);
   }
  } /* Quantization and coding block */

}


void couple_e(DSPfract *cplmix, DSPshort nfchans, float *cpltc)
{
	int bin, chan;
	double invcplmix;

	if ((cplmix[0] > 0.01) || (cplmix[0] < -0.01))
	{
		for (bin = 0; bin < block_size_samples; bin++)
		{
			cpltc[bin] *= cplmix[0];
		}

		invcplmix = 1.0 / cplmix[0];
		for (chan = 1; chan < nfchans; chan++)
		{
			cplmix[chan] *= invcplmix;
		}
	}
	else
	{
		CommonExit(1, "Must not exceed 40dB attenuation in the left channel\n");
	}
}












