/*******************************************************************************
This software module was originally developed by

Institute for Infocomm Research and Fraunhofer IIS

and edited by

-

in the course of development of ISO/IEC 14496 for reference purposes and its
performance may not have been optimized. This software module is an
implementation of one or more tools as specified by ISO/IEC 14496. ISO/IEC gives
You a royalty-free, worldwide, non-exclusive, copyright license to copy,
distribute, and make derivative works of this software module or modifications
thereof for use in implementations of ISO/IEC 14496 in products that satisfy
conformance criteria (if any). Those intending to use this software module in
products are advised that its use may infringe existing patents. ISO/IEC have no
liability for use of this software module or modifications thereof. Copyright is
not released for products that do not conform to audiovisual and image-coding
related ITU Recommendations and/or ISO/IEC International Standards.

#ifdef NOT_PUBLISHED

Assurance that the originally developed software module can be used (1) in
ISO/IEC 14496 once ISO/IEC 14496 has been adopted; and (2) to develop ISO/IEC
14496:
Institute for Infocomm Research and Fraunhofer IIS grant ISO/IEC all
rights necessary to include the originally developed software module or
modifications thereof in ISO/IEC 14496 and to permit ISO/IEC to offer You a
royalty-free, worldwide, non-exclusive, copyright license to copy, distribute,
and make derivative works for use in implementations of ISO/IEC 14496 in
products that satisfy conformance criteria (if any), and to the extent that such
originally developed software module or portions of it are included in ISO/IEC
14496. To the extent that Institute for Infocomm Research and Fraunhofer IIS 
own patent rights that would be required to make, use, or sell the
originally developed software module or portions thereof included in ISO/IEC
14496 in a conforming product, Institute for Infocomm Research and Fraunhofer IIS 
will assure the ISO/IEC that they are willing to negotiate
licenses under reasonable and non-discriminatory terms and conditions with
applicants throughout the world. ISO/IEC gives You a free license to this
software module or modifications thereof for the sole purpose of developing
ISO/IEC 14496.

#endif

Institute for Infocomm Research and Fraunhofer IIS retain full right to
modify and use the code for their own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for products
that do not conform to MPEG-related ITU Recommendations and/or ISO/IEC
International Standards. This copyright notice must be included in all copies or
derivative works.

Copyright (c) ISO/IEC 2005.
*******************************************************************************/



#include <stdio.h>

#include "int_mdct.h"


static void swapInt(int* x, int* y)
{
  int dummy;

  dummy = *x;
  *x = *y;
  *y = dummy;
}


/* IntMDCT:
   Processed buffer has length N+N/2. 
   The first N/2 values contain TDA+windowed signal from last frame.
   The next N signal contain new frame of time domain input signal.
   TDA+windowing is done on this length N buffer. The right N/2 should be 
   stored for next frame.
   DCT-IV is performed on left N TDA+w samples.
   So MDCT output is stored in the left N samples.

   
   < N/2 TDA+w > <     N time samples      >

                              | 
                            TDA+w
                              |

   < N/2 TDA+w > < N/2 TDA+w > < N/2 TDA+w >

                |
       minus mirrored DCT-IV
                |

   <      N MDCT lines       > < N/2 TDA+w >

*/
void IntMDCT(int* p_in,
	     int* p_overlap,
	     int* p_out,
	     int blockType,
	     int windowShape,
	     int osf)
{
  int i;
  int j;
  int N;
  int transFac = 8;
  int NShort;
  int startOnesLength; /* number of ones in start window */
  int signal[MAX_OSF*INTMDCT_BUFFER_SIZE];

  N = 1024*osf;
  NShort = N/transFac;
  startOnesLength = N/2 - NShort/2;

  /* copy buffers */
  for (i=0; i<N/2; i++) {
    signal[i]=p_overlap[i];
  }
  for (i=0; i<N; i++) {
    signal[N/2+i]=p_in[i];
  }
  
  /* windowing and TDA */
  switch(blockType){
  case ONLY_LONG_WINDOW:
  case SHORT_LONG_WINDOW:
    /* perform TDA and windowing of right N samples 
       based on Integer Rotations */ 
    windowingTDA(&signal[N/2],
		 N,
		 windowShape,
		 1);
    break;
  case LONG_SHORT_WINDOW:
    windowingTDA(&signal[N/2+startOnesLength],
		 NShort,
		 windowShape,
		 1);
    break;
  case ONLY_SHORT_WINDOW:
    for (j=0; j<transFac; j++) {
      windowingTDA(&signal[NShort/2+j*NShort],
		   NShort,
		   windowShape,
		   1);
    }
    break;
  default:
    printf("blockType not implemented!\n");
    break;
  }
    
  /* DCT-IV */
  switch(blockType){
  case ONLY_LONG_WINDOW:
  case LONG_SHORT_WINDOW:
  case SHORT_LONG_WINDOW:
  
    /* revert order of N samples to process */
    for (i=0; i<N/2; i++) {
      swapInt(&signal[i],&signal[N-1-i]);
    }
    
    for (i=0; i<N; i++) {
      signal[i] *= -1;
    }

    /* perform Integer DCT-IV on left N samples */
    intDCTIV(signal,
	     &signal[N/2],
	     N,
	     1);
    break;
  case ONLY_SHORT_WINDOW:
    for (j=0; j<transFac; j++) {
      /* revert order of N samples to process */
      for (i=0; i<NShort/2; i++) {
	swapInt(&signal[j*NShort+i],&signal[j*NShort+NShort-1-i]);
      }
      for (i=0; i<NShort; i++) {
	signal[j*NShort+i] *= -1;
      }

      intDCTIV(&signal[j*NShort],
	       &signal[j*NShort+NShort/2],
	       NShort,
	       1);
    }
    break;  
  default:
    printf("blockType not implemented!\n");
    break;
  }

  /* copy buffers */
  for (i=0; i<N; i++) {
    p_out[i]=signal[i];
  }
  for (i=0; i<N/2; i++) {
    p_overlap[i]=signal[N+i];
  }

}


void StereoIntMDCT(int* p_in_0,
		   int* p_in_1,
		   int* p_overlap_0,
		   int* p_overlap_1,
		   int* p_out_0,
		   int* p_out_1,
		   int blockType,
		   int windowShape,
		   int osf)
{
  int i;
  int j;
  int N;
  int transFac = 8;
  int NShort;
  int startOnesLength; /* number of ones in start window */
  int signal[2][MAX_OSF*INTMDCT_BUFFER_SIZE];
  int ch;

  N = 1024*osf;
  NShort = N/transFac;
  startOnesLength = N/2 - NShort/2;

  /* copy buffers */
  for (i=0; i<N/2; i++) {
    signal[0][i]=p_overlap_0[i];
    signal[1][i]=p_overlap_1[i];
  } 
  for (i=0; i<N; i++) {
    signal[0][N/2+i]=p_in_0[i];
    signal[1][N/2+i]=p_in_1[i];
  }

  for (ch=0; ch<2; ch++) {
    /* windowing and TDA */
    switch(blockType){
    case ONLY_LONG_WINDOW:
    case SHORT_LONG_WINDOW:
      /* perform TDA and windowing of right N samples 
	 based on Integer Rotations */ 
      windowingTDA(&signal[ch][N/2],
		   N,
		   windowShape,
		   1);
      break;
    case LONG_SHORT_WINDOW:
      windowingTDA(&signal[ch][N/2+startOnesLength],
		   NShort,
		   windowShape,
		   1);
      break;
    case ONLY_SHORT_WINDOW:
      for (j=0; j<transFac; j++) {
	windowingTDA(&signal[ch][NShort/2+j*NShort],
		     NShort,
		     windowShape,
		     1);
      }
      break;
    default:
      printf("blockType not implemented!\n");
      break;
    }
  }

  for (ch=0; ch<2; ch++) {
    /* DCT-IV */
    switch(blockType){
    case ONLY_LONG_WINDOW:
    case LONG_SHORT_WINDOW:
    case SHORT_LONG_WINDOW:
  
      /* revert order of N samples to process */
      for (i=0; i<N/2; i++) {
	swapInt(&signal[ch][i],&signal[ch][N-1-i]);
      }
    
      for (i=0; i<N; i++) {
	signal[ch][i] *= -1;
      }
      break;
    case ONLY_SHORT_WINDOW:
      for (j=0; j<transFac; j++) {
	/* revert order of N samples to process */
	for (i=0; i<NShort/2; i++) {
	  swapInt(&signal[ch][j*NShort+i],&signal[ch][j*NShort+NShort-1-i]);
	}
	for (i=0; i<NShort; i++) {
	  signal[ch][j*NShort+i] *= -1;
	}
      }
      break;  
    default:
      printf("blockType not implemented!\n");
      break;
    }
  }

  /* DCT-IV */
  switch(blockType){
  case ONLY_LONG_WINDOW:
  case LONG_SHORT_WINDOW:
  case SHORT_LONG_WINDOW:
    
    /* perform Integer DCT-IV on left N samples */
    intDCTIV(signal[0],
	     signal[1],
	     2*N,
	     0);
    break;
  case ONLY_SHORT_WINDOW:
    for (j=0; j<transFac; j++) {
      intDCTIV(&signal[0][j*NShort],
	       &signal[1][j*NShort],
	       2*NShort,
	       0);
    }
    break;  
  default:
    printf("blockType not implemented!\n");
    break;
  }

  /* copy buffers */
  for (i=0; i<N; i++) {
    p_out_0[i]=signal[0][i];
    p_out_1[i]=signal[1][i];
  }
  for (i=0; i<N/2; i++) {
    p_overlap_0[i]=signal[0][N+i];
    p_overlap_1[i]=signal[1][N+i];
  }
}









